import pytest

import operator

import datamol as dm
import numpy as np

from lark import Lark, exceptions

from medchem.query import QueryParser
from medchem.query import QueryOperator
from medchem.query import QueryFilter
from medchem.query import EvaluableQuery

from medchem.utils.loader import get_grammar


def get_test_grammar():
    return Lark(get_grammar(as_string=True), parser="lalr", transformer=QueryParser())


def test_language():
    grammar = get_test_grammar()

    bad_queries = [
        "HASPROP(tpsa > 120 )",  # missing quotes, UnexpectedCharacters
        """HASPROP("tpsa" > 120 ) OR OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C")""",  # wrong syntax with 2 OR, UnexpectedCharacters
        """HASPROP("tpsa" > 120 ) OR TRUE("wrong")""",  # wrong syntax, UnexpectedCharacters
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C") OR""",  # wrong syntax, wrong terminal node (OR), KeyError
        """HASPROP("tpsa" > 120 ) OR  (HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C")""",  # wrong syntax, parenthesis mismatch, KeyError
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C") TRUE""",  # wrong syntax, wrong terminal node (TRUE) , KeyError
        """HASPROP("tpsa" > 120 ) OR () OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C")""",  # wrong syntax, empty statement, UnexpectedCharacters
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True, >, 3)""",  # unexpected token '>' at wrong place, UnexpectedCharacters
        """HASPROP("tpsa" + 120) OR HASSUBSTRUCTURE("[OH]")"""  # unexpected token '+', UnexpectedCharacters
        """HASPROP("tpsa") OR HASSUBSTRUCTURE("[OH]")""",  # incomplete HASPROP statement UnexpectedCharacters
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]" True max 3)""",  # mixing comma in substructure statement token, UnexpectedCharacters
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", min, True, 3)""",  # wrong order of arguments, UnexpectedCharacters
        """HASPROP("tpsa" > 120) OR BIND("herg")""",  # BIND is not a valid statement
        """HASALERT("tpsa" > 120)""",  # this is not a valid statement
    ]
    valid_queries = [
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("ls -lah /")""",  # This is valid query, but "ls -lah" will never be evaluated
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C")""",
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C") OR NOT TRUE""",  # this is valid, but should be simplified,
        """HASPROP("tpsa" > 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C") OR TRUE""",
        """HASPROP("tpsa",  >, 120 ) OR HASSUPERSTRUCTURE("Cc1cc2ccccc2cc1C")""",
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]")""",  # min match, not smarts, limit=1
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True)""",  # min match, not smarts, limit=1
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True, min)""",  # min match, not smarts, limit=1
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True, "min")""",  # doesn't matter if min is quoted
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True, min, 1)""",  # min match, not smarts, limit=1
        """HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("[OH]", True, 1)""",  # this is correct but bad practice
    ]
    for query in bad_queries:
        with pytest.raises((exceptions.UnexpectedToken, KeyError, exceptions.UnexpectedCharacters)):
            print(query)
            grammar.parse(query)

    valid_results = [grammar.parse(query) is not None for query in valid_queries]
    assert valid_results == [True] * len(valid_queries)


def test_query_equivalence():
    grammar = get_test_grammar()

    # the following should all be equal in the resulting query
    equivalent_queries = [
        """HASPROP("tpsa" = 120) OR ! HASSUBSTRUCTURE("CO")""",  # A OR ! B
        """HASPROP("tpsa", =, 120) OR ! HASSUBSTRUCTURE("CO")""",  # A OR ! B, comma does not matter
        """HASPROP("tpsa" =, 120) OR ! HASSUBSTRUCTURE("CO")""",  # A OR ! B, comma does not matter
        'HASPROP("tpsa" = 120) OR ! HASSUBSTRUCTURE("CO")',  # A OR ! B
        """(HASPROP("tpsa" = 120)) OR ! (HASSUBSTRUCTURE("CO"))""",  # (A) OR  ! (B)
        """(HASPROP("tpsa" = 120) OR ! HASSUBSTRUCTURE("CO"))""",  # (A OR ! B)
        """HASPROP("tpsa" = 120) OR NOT HASSUBSTRUCTURE("CO")""",  # A OR NOT B
        """HASPROP("tpsa" = 120) OR ~HASSUBSTRUCTURE("CO")""",  # A OR ~B
        """HASPROP("tpsa" == 120) OR ! HASSUBSTRUCTURE("CO")""",  # A OR ! B
        """HASPROP("tpsa" =   120)   OR ! HASSUBSTRUCTURE("CO")""",  # change in space
    ]
    different_query = """( ! HASPROP("tpsa" > 120) OR HASSUBSTRUCTURE("CO"))"""
    same_results = [grammar.parse(equivalent_queries[i]) for i in range(len(equivalent_queries))]

    expected_equivalent_results = [same_results[i] == same_results[0] for i in range(len(same_results))]
    expected_different_results = grammar.parse(different_query)

    assert expected_equivalent_results == [True] * len(same_results)
    assert expected_different_results != same_results[0]


def test_prop():
    mol = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")
    tpsa = dm.descriptors.tpsa(mol)
    clogp = dm.descriptors.clogp(mol)

    assert QueryOperator.getprop(mol, "tpsa") == tpsa
    assert QueryOperator.getprop(mol, "clogp") == clogp

    with pytest.raises(Exception):
        QueryOperator.getprop(mol, "fake")

    with pytest.raises(Exception):
        QueryOperator.hasprop(mol, "tpsa", ">", 120)  # type: ignore

    assert QueryOperator.hasprop(mol, "tpsa", operator.le, 120) is True


def test_alert():
    mol = dm.to_mol("Oc1cscc1-c1ccc(O)cc1")  # contains tiophene hydroxy should match pains

    assert QueryOperator.hasalert(mol, "pains") is True
    assert QueryOperator.hasalert(mol, "brenk") is False

    with pytest.raises(Exception):
        QueryOperator.hasalert(mol, "fake")


def test_substructure():
    mol = "Oc1cscc1-c1ccc(OC)cc1"  # contains tiophene hydroxy should match pains
    substruct1 = "s1ccc(c1)-[#8;H1]"
    substruct2 = "CO"  # should match twice for SMILES and once for SMARTS

    assert QueryOperator.hassubstructure(mol, substruct1, is_smarts=True) is True
    assert QueryOperator.hassubstructure(mol, substruct2, is_smarts=True) is True
    assert QueryOperator.hassubstructure(mol, substruct2, is_smarts=False) is True
    assert QueryOperator.hassubstructure(mol, substruct2, True, "min", limit=1) is True
    assert QueryOperator.hassubstructure(mol, substruct2, True, "min", limit=2) is False
    assert QueryOperator.hassubstructure(mol, substruct2, True, None, limit=2) is False
    assert QueryOperator.hassubstructure(mol, substruct2, True, "max", limit=2) is True
    assert QueryOperator.hassubstructure(mol, substruct2, False, "min", limit=1) is True
    assert QueryOperator.hassubstructure(mol, substruct2, False, "min", limit=2) is True
    assert QueryOperator.hassubstructure(mol, substruct2, False, None, limit=2) is True
    assert QueryOperator.hassubstructure(mol, substruct2, False, "max", limit=2) is False


def test_rule():
    mol = "Oc1cscc1-c1ccc(OC)cc1"
    assert QueryOperator.matchrule(mol, "rule_of_five") is True
    assert QueryOperator.matchrule(mol, "rule_of_three") is False

    with pytest.raises(Exception):
        QueryOperator.matchrule(mol, "fake")


def test_group():
    mol = "Oc1cscc1-c1ccc(OC)cc1"
    assert QueryOperator.hasgroup(mol, "Ethers") is True
    assert QueryOperator.hasgroup(mol, "Ketones") is False

    # should normally match an alcohol, but how the group as defined, the specificity is high
    # so another more specific group should match and not alcohol for the tiophene hydroxy
    assert QueryOperator.hasgroup(mol, "Alcohols") is False
    assert QueryOperator.hasgroup(mol, "Hydroxy compounds: alcohols or phenols") is True

    with pytest.raises(Exception):
        QueryOperator.hasgroup(mol, "fake")


def test_similarity():
    mol1 = "Oc1cscc1-c1ccc(OC)cc1"
    mol2 = dm.to_mol("Oc1cscc1-c1ccc(O)cc1")
    dist = dm.cdist([mol1], [mol2])[0][0]

    assert QueryOperator.like(mol1, mol2, operator.le, 0.8) == True  # noqa: E712
    assert QueryOperator.similarity(mol1, mol2) == 1 - dist


def test_superstructure():
    query = "Oc1cscc1-c1ccc(OC)cc1"
    mol = "Oc1ccsc1"

    assert QueryOperator.hassuperstructure(mol, query) is True
    assert QueryOperator.hassuperstructure(query, mol) is False


def test_query_eval():
    mol = "Oc1cscc1-c1ccc(OC)cc1"
    grammar = Lark(get_grammar(as_string=True), parser="lalr", transformer=QueryParser())
    queries = [
        """HASPROP("tpsa" <= 120) AND ! HASALERT("pains") AND HASGROUP("Alcohols")""",  # False, does not match alchol
        """HASPROP("tpsa" <= 120) AND NOT HASALERT("pains") AND HASGROUP("Ethers")""",  # False has pains alerts
        """(HASPROP("tpsa" <= 120) AND NOT HASALERT("brenk")) AND HASGROUP("Ethers")""",  # True
        """HASPROP("tpsa" <= 120) AND MATCHRULE("rule_of_three")""",  # False
        """HASPROP("tpsa" <= 120) AND HASSUBSTRUCTURE("[OH]", True)""",  # True
        """HASPROP("tpsa" <= 120) AND HASSUBSTRUCTURE("[OH]", True, 3)""",  # False
    ]
    parsed_queries = [grammar.parse(q) for q in queries]
    expected_results = [False, False, True, False, True, False]

    observed_results = []
    observed_results_str = []
    for q in parsed_queries:
        evaluator = EvaluableQuery(q)
        observed_results.append(evaluator(mol))
        observed_results_str.append(evaluator(mol, exec=False))

    assert expected_results == observed_results


def test_query_filter():
    queries = [
        # complex query
        """(HASPROP("tpsa" < 100) AND HASPROP("clogp" < 3) AND ! HASALERT("pains")) OR (HASPROP("n_heavy_atoms" >= 10) AND (HASGROUP("Alcohols") OR HASSUBSTRUCTURE("[CX3](=[OX1])O", True, 1)))""",
        # is a rewriting of the above
        """(HASPROP("tpsa" < 100) AND HASPROP("clogp" < 3) AND ! HASALERT("pains")) OR (HASPROP("n_heavy_atoms" >= 10) AND (HASGROUP("Alcohols") OR HASSUBSTRUCTURE("[CX3](=[OX1])O", True, min, 1)))""",
        # is a rewriting of the above with spacing and a differentm yet equivalent bool expression
        """
        (
            HASPROP("tpsa" < 100)
            AND
            HASPROP("clogp" < 3)
            AND
            ! HASALERT("pains")
        )
        OR
        (
            HASPROP("n_heavy_atoms" >= 10)
            AND
            HASGROUP("Alcohols")
        )
        OR
        (
            HASPROP("n_heavy_atoms" >= 10)
            AND
            HASSUBSTRUCTURE("[CX3](=[OX1])O", True)
        )
        """,
        # always true
        """(HASPROP("tpsa" < 100) AND HASPROP("clogp" < 3)) OR TRUE""",
        # always false
        """(HASPROP("tpsa" < 100) AND HASPROP("clogp" < 3)) AND False""",
    ]

    query_filters = [QueryFilter(q) for q in queries]
    df = dm.cdk2()
    df["tpsa"] = df["mol"].apply(dm.descriptors.tpsa)
    df["clogp"] = df["mol"].apply(dm.descriptors.clogp)
    df["n_heavy_atoms"] = df["mol"].apply(dm.descriptors.n_heavy_atoms)
    df["has_carboxyl"] = df["mol"].apply(lambda x: QueryOperator.hassubstructure(x, "[CX3](=[OX1])O", True))
    df["has_pains"] = df["mol"].apply(lambda x: QueryOperator.hasalert(x, "pains"))
    df["has_alcohol"] = df["mol"].apply(lambda x: QueryOperator.hasgroup(x, "Alcohols"))

    tmp = df.query(
        "((tpsa < 100) & (clogp < 3) & ~has_pains) | (n_heavy_atoms >= 10 & (has_carboxyl | has_alcohol))"
    )
    df["expected_results"] = False
    df.loc[tmp.index, "expected_results"] = True

    computed_1 = query_filters[0](df["mol"].tolist(), progress=False)
    computed_2 = query_filters[1](df["mol"].tolist(), progress=False)
    computed_3 = query_filters[2](df["smiles"].tolist(), progress=False)
    all_true = query_filters[3](df["mol"].tolist(), progress=False)
    all_false = query_filters[4](df["mol"].tolist(), progress=False)

    np.testing.assert_array_equal(np.asarray(computed_1), np.asarray(computed_2))
    np.testing.assert_array_equal(np.asarray(computed_1), np.asarray(computed_3))
    np.testing.assert_array_equal(df.expected_results, np.asarray(computed_1))
    np.testing.assert_array_equal(np.asarray(all_true), [True] * len(df))
    np.testing.assert_array_equal(np.asarray(all_false), [False] * len(df))



import typer

from loguru import logger

import medchem as mc
import datamol as dm
import pandas as pd

app = typer.Typer(help="The Medchem CLI", add_completion=False)


@app.command(help="Filtering for common structural alerts")
def common_alerts(
    input: str = typer.Argument(..., help="Input file path (CSV, Excel, Parquet, JSON, SDF)"),
    output: str = typer.Argument(..., help="Output file path (CSV, Excel, Parquet, JSON, SDF)"),
    smiles_column: str = typer.Option("smiles", help="The name of the column containing the SMILES"),
    n_jobs: int = typer.Option(-1, help="Number of jobs to use"),
    scheduler: str = typer.Option("auto", help="The scheduler to use"),
    keep_details: bool = typer.Option(False, help="Whether to keep the details or not"),
    progress: bool = typer.Option(True, help="Whether to show progress or not"),
    progress_leave: bool = typer.Option(False, help="Whether to leave the progress bar or not"),
):
    logger.info(f"Loading data from {input}")
    data = dm.io.open_df(input)

    logger.info(f"Filtering {data.shape[0]} molecules")
    filters_obj = mc.structural.CommonAlertsFilters()

    results = filters_obj(
        mols=data[smiles_column].tolist(),
        n_jobs=n_jobs,
        scheduler=scheduler,
        keep_details=keep_details,
        progress=progress,
        progress_leave=progress_leave,
    )

    results = pd.concat([data[[smiles_column]], results.drop(columns="mol")], axis=1)

    logger.info(f"Saving results to {output}")
    dm.io.save_df(results, output)


@app.command(help="Filtering with NIBR structural alerts")
def nibr_filters(
    input: str = typer.Argument(..., help="Input file path (CSV, Excel, Parquet, JSON, SDF)"),
    output: str = typer.Argument(..., help="Output file path (CSV, Excel, Parquet, JSON, SDF)"),
    smiles_column: str = typer.Option("smiles", help="The name of the column containing the SMILES"),
    n_jobs: int = typer.Option(-1, help="Number of jobs to use"),
    scheduler: str = typer.Option("threads", help="The scheduler to use"),
    keep_details: bool = typer.Option(False, help="Whether to keep the details or not"),
    progress: bool = typer.Option(True, help="Whether to show progress or not"),
    progress_leave: bool = typer.Option(False, help="Whether to leave the progress bar or not"),
):
    logger.info(f"Loading data from {input}")
    data = dm.io.open_df(input)

    logger.info(f"Filtering {data.shape[0]} molecules")
    filters_obj = mc.structural.NIBRFilters()

    results = filters_obj(
        mols=data[smiles_column].tolist(),
        n_jobs=n_jobs,
        scheduler=scheduler,
        keep_details=keep_details,
        progress=progress,
        progress_leave=progress_leave,
    )

    results = pd.concat([data[[smiles_column]], results.drop(columns="mol")], axis=1)

    logger.info(f"Saving results to {output}")
    dm.io.save_df(results, output)


if __name__ == "__main__":
    app()  # pragma: no cover



import pytest


import datamol as dm

from medchem.constraints import Constraints


def test_constraints():
    def my_constraint(mol):
        # we want to either (have phenol) OR (have less than 7 atoms and not ring)
        return mol.HasSubstructMatch(dm.to_mol("Oc1ccccc1")) or (
            mol.GetNumAtoms() < 7 and dm.descriptors.n_rings(mol) < 1
        )

    smiles = [
        "CN(C)C(=O)c1cncc(C)c1",  # match, n_atoms < 7 and no ring
        "Cc1cncc(CC2CCCCC2)c1",  # not match, n_atoms < 7 but ring
        "Cc1cncc(c1)-c1ccc(O)cc1",  # match phenol
        "Cc1cncc(c1)-c1cccc2nc[nH]c12",  # no match n_atoms >= 7
    ]
    expected_results = []
    mols = [dm.to_mol(x) for x in smiles]
    core = dm.from_smarts("[C;H3]c1cncc([*:1])c1")

    assert core is not None

    # now let's set the constraints query
    for atom in core.GetAtoms():
        # we add a recursive query to check again on any match that starts with this atom position
        if atom.GetAtomMapNum() == 1:
            atom.SetProp("query", "my_constraint")

    constraint_fns = {"my_constraint": my_constraint}
    constraint = Constraints(core, constraint_fns)

    matches = [constraint(mol) for mol in mols]
    expected_results = [True, False, True, False]
    assert expected_results == matches

    one_match = constraint.get_matches(mols[0], multiple=False)
    assert isinstance(one_match[0], int)
    assert len(one_match) > 0


def test_constraints_validate():
    def my_constraint(mol):
        return mol.HasSubstructMatch(dm.to_mol("Oc1ccccc1")) or (
            mol.GetNumAtoms() < 7 and dm.descriptors.n_rings(mol) < 1
        )

    core = dm.from_smarts("[C;H3]c1cncc([*:1])c1")

    assert core is not None

    for atom in core.GetAtoms():
        if atom.GetAtomMapNum() == 1:
            atom.SetProp("query", "my_constraint")

    # Create the constraints object
    constraint_fns = {"my_constraint": my_constraint}
    constraint = Constraints(core, constraint_fns)

    mol = dm.to_mol("CN(C)C(=O)c1cncc(C)c1")

    assert Constraints.validate(mol, constraints=[constraint]) is True


def test_constraints_validate_fails():
    mol = dm.to_mol("CN(C)C(=O)c1cncc(C)c1")

    with pytest.raises(ValueError):
        Constraints.validate(mol, constraints=["xxxxxx"])  # type: ignore

    with pytest.raises(ValueError):
        Constraints.validate(None, constraints=[])  # type: ignore



import datamol as dm
import medchem as mc


def test_alert_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]

    results = mc.functional.alert_filter(
        mols=data["smiles"].tolist(),
        alerts=["BMS"],
        n_jobs=-1,
        progress=True,
        return_idx=False,
    )

    assert results.tolist() == [True, False, True, True, True, True, True, True, True, True]

    results = mc.functional.alert_filter(
        mols=data["smiles"].tolist(),
        alerts=["BMS"],
        n_jobs=-1,
        progress=True,
        return_idx=True,
    )

    assert results.tolist() == [0, 2, 3, 4, 5, 6, 7, 8, 9]


def test_nibr_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]

    results = mc.functional.nibr_filter(
        mols=data["smiles"].tolist(),
        n_jobs=-1,
        progress=True,
        return_idx=False,
    )

    assert results.tolist() == [True, False, True, True, True, True, True, True, True, True]

    results = mc.functional.nibr_filter(
        mols=data["smiles"].tolist(),
        n_jobs=-1,
        progress=True,
        return_idx=True,
    )

    assert results.tolist() == [0, 2, 3, 4, 5, 6, 7, 8, 9]


def test_catalog_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]

    results = mc.functional.catalog_filter(
        mols=data["smiles"].tolist(),
        catalogs=["tox", "pains"],
        n_jobs=-1,
        progress=True,
        progress_leave=True,
        return_idx=False,
        batch_size=100,
        scheduler="threads",
    )

    assert results.tolist() == [True, True, False, True, False, True, True, True, True, True]

    results = mc.functional.catalog_filter(
        mols=data["smiles"].tolist(),
        catalogs=["tox", "pains"],
        n_jobs=-1,
        progress=True,
        progress_leave=True,
        return_idx=True,
        batch_size=100,
        scheduler="threads",
    )

    assert results.tolist() == [0, 1, 3, 5, 6, 7, 8, 9]


def test_chemical_group_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]

    cg = mc.groups.ChemicalGroup("common_organic_solvents")

    results = mc.functional.chemical_group_filter(
        mols=data["smiles"].tolist(),
        chemical_group=cg,
        n_jobs=-1,
        progress=True,
        progress_leave=True,
        return_idx=False,
        scheduler="threads",
    )

    assert results.tolist() == [False, True, False, True, False, False, True, False, False, False]

    results = mc.functional.chemical_group_filter(
        mols=data["smiles"].tolist(),
        chemical_group=cg,
        n_jobs=-1,
        progress=True,
        progress_leave=True,
        return_idx=True,
        scheduler="threads",
    )

    assert results.tolist() == [1, 3, 6]


def test_rules_filter():
    rfilter = mc.rules.RuleFilters(rule_list=["rule_of_five", "rule_of_oprea", "rule_of_cns"])

    data = dm.data.cdk2()
    data = data.iloc[:10]

    results = mc.functional.rules_filter(
        mols=data["smiles"].tolist(),
        rules=rfilter,
        return_idx=False,
    )

    assert results.tolist() == [False, False, False, True, True, False, False, False, False, True]

    results = mc.functional.rules_filter(
        mols=data["smiles"].tolist(),
        rules=rfilter,
        return_idx=True,
    )
    assert results.tolist() == [3, 4, 9]


def test_complexity_filter():
    data = dm.data.cdk2()
    data = data.iloc[:10]

    results = mc.functional.complexity_filter(
        mols=data["smiles"].tolist(),
        return_idx=False,
    )

    assert results.tolist() == [True, True, True, True, True, True, True, True, True, True]

    results = mc.functional.complexity_filter(
        mols=data["smiles"].tolist(),
        return_idx=True,
    )

    assert results.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_bredt_filter():
    bredt_test_set = [
        "C1C2=C1C2",
        "C1CC=C=CC1",
        "C1C2CCCC=C12",
        "C1C2=C1CCCC2",  # is ok
        "C1CC2=CCC1C2",
        "C1CC2=CC1CC2",
        "C1CC2CCC1C=C2",  # is ok
        "C1CC2=CCC1CC2",
    ]

    results = mc.functional.bredt_filter(mols=bredt_test_set, return_idx=False)

    assert results.tolist() == [False, False, False, True, False, False, True, False]

    results = mc.functional.bredt_filter(mols=bredt_test_set, return_idx=True)

    assert results.tolist() == [3, 6]


def test_molecular_graph_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]

    results = mc.functional.molecular_graph_filter(mols=data["smiles"].tolist(), return_idx=False)

    assert results.tolist() == [True, True, True, True, True, True, True, True, True, True]

    results = mc.functional.molecular_graph_filter(mols=data["smiles"].tolist(), return_idx=True)

    assert results.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_lilly_demerit_filter():
    data = dm.data.cdk2()
    data = data.iloc[:10]

    results = mc.functional.lilly_demerit_filter(mols=data["smiles"].tolist(), return_idx=True)

    assert results.tolist() == [0, 1, 2, 3, 4, 5, 6, 7]

    results = mc.functional.lilly_demerit_filter(mols=data["smiles"].tolist(), return_idx=False)
    assert results.tolist() == [True, True, True, True, True, True, True, True, False, False]


def test_protecting_groups_filter():
    non_protected_smiles = [
        "C1=CC=C2C(=C1)C(=O)OC2(C3=CC=C(C=C3)O)C4=CC=C(C=C4)O",
        "CC(=CCCC(C)(C=C)OC(=O)C)C",
        "CCC1(C(=O)N(C(=O)O1)C)C",
        "CC1=CCC(CC1)C(C)(C)OC(=O)C",
        "CCC(=O)OC(C)(C)C1CCC(=CC1)C",
    ]
    protected_smiles = [
        "CC(C)(C)OC(=O)N1CCC(CC1)N",
        "CC(C)(C)OC(=O)NCC(=O)O",
        "CC(C)(C)OC(=O)N1CCCC1C(=O)O",
        "CC(C)(C)OC(=O)NC(COCC1=CC=CC=C1)C(=O)O",
        "CC(=O)OC(C)(C)C",
    ]
    non_protected_mols = [dm.to_mol(x) for x in non_protected_smiles]
    protected_mols = [dm.to_mol(x) for x in protected_smiles]
    data = non_protected_mols + protected_mols
    results = mc.functional.protecting_groups_filter(data, return_idx=False)

    assert results.tolist() == [True, True, True, True, True, False, False, False, False, False]

    results = mc.functional.protecting_groups_filter(mols=data, return_idx=True)

    assert results.tolist() == [0, 1, 2, 3, 4]



from typing import List
from typing import Optional
from typing import Union
from typing import Sequence

import os
import functools

import pandas as pd
import datamol as dm

from loguru import logger

from medchem.utils.loader import get_data_path


class CommonAlertsFilters:
    """Filtering class for building a library based on a list of structural alerts

    To list the available alerts, use the `list_default_available_alerts` method.

    The output of the filter are explained below:
    - **status**: one of `["exclude", "flag", "annotations", "ok"]` (ordered by quality).
        Generally, you can keep anything without the "exclude" label, as long as you also apply
        a maximum severity score for compounds that collects too many flags.
    - **reasons**: list of reasons why the compound was flagged.
    - **pass_filter**: whether the compound passed the filter or not.
    - **details**: optional additional details of the evaluation, including matching patterns if `keep_details` is True.
    """

    def __init__(
        self,
        alerts_set: Optional[Union[str, List[str]]] = None,
        alerts_db_path: Optional[Union[os.PathLike, str]] = None,
    ):
        """Filtering molecules based on chemical alerts

        Args:
            alerts_set: Filter set to use. Default is BMS+Dundee+Glaxo.
            alerts_db_path: Alerts file to use. Default is internal.
        """
        if alerts_db_path is None:
            alerts_db_path = get_data_path(filename="common_alerts_collection.csv")

        if alerts_set is None:
            alerts_set = ["BMS"]
        elif isinstance(alerts_set, str):
            alerts_set = [alerts_set]

        alerts_set = [x.lower() for x in set(alerts_set)]

        self.alerts_df = self._build_alerts(alerts_set=alerts_set, alerts_db_path=alerts_db_path)

    def _build_alerts(
        self,
        alerts_set: Union[str, List[str]],
        alerts_db_path: Union[os.PathLike, str],
    ):
        # Load the db
        alerts_df = pd.read_csv(alerts_db_path)

        # Select only the filters we want
        mask = alerts_df["rule_set_name"].str.lower().isin(alerts_set)
        alerts_df = alerts_df[mask]
        alerts_df = alerts_df.reset_index(drop=True)

        with dm.without_rdkit_log():
            alerts_df["smarts_mol"] = alerts_df["smarts"].apply(dm.from_smarts)

        # Check for invalid SMARTS
        mask_invalid = alerts_df["smarts_mol"].isnull()
        if mask_invalid.sum() > 0:
            invalid_smarts = alerts_df.loc[mask_invalid, "smarts"].tolist()  # pragma: no cover
            logger.warning(
                f"The following SMARTS are invalid and will be ignored: {invalid_smarts}"
            )  # pragma: no cover

        return alerts_df

    @staticmethod
    @functools.lru_cache()
    def list_default_available_alerts():
        """
        Return a list of unique alert set names
        """
        alerts_db = get_data_path(filename="common_alerts_collection.csv")
        rule_df = pd.read_csv(alerts_db)
        rule_list = (
            rule_df.groupby("rule_set_name")
            .agg(
                {
                    "smarts": "count",
                    "catalog_description": "first",
                    "rule_set": "first",
                    "source": "first",
                }
            )
            .sort_values("rule_set")
            .reset_index()
        )

        return rule_list

    def _evaluate(
        self,
        mol: Union[str, dm.Mol],
        keep_details: bool = False,
    ):
        """
        Evaluate structural alerts on a molecule

        Args:
            mol: input molecule
        """
        if isinstance(mol, str):
            with dm.without_rdkit_log():
                _mol = dm.to_mol(mol)
        else:
            _mol = mol

        if _mol is None:
            return pd.Series(
                {
                    "mol": _mol,
                    "pass_filter": False,
                    "status": "exclude",
                    "reasons": "invalid",
                }
            )

        # Match the molecule against the alerts
        matches = self.alerts_df["smarts_mol"].apply(lambda x: _mol.GetSubstructMatches(x))
        n_matches = matches.apply(len)
        has_alert = n_matches > 0

        # Prepare the result
        result = pd.Series()
        result["mol"] = _mol

        if any(has_alert):
            result["pass_filter"] = False
            result["status"] = "exclude"
            result["reasons"] = ";".join(self.alerts_df[has_alert]["description"].dropna().unique().tolist())

            # NOTE(hadim): we could eventually backpropagate the full alert definition of the matching alerts +
            # match atom indices and also the molecule object with the highlighted atoms.

        else:
            result["pass_filter"] = True
            result["status"] = "ok"
            result["reasons"] = None

        if keep_details:
            details = self.alerts_df[has_alert].copy()
            details = details.drop(columns=["smarts_mol"])
            details["matches"] = matches[has_alert].tolist()
            result["details"] = details.to_dict(orient="records")

        return result

    def __call__(
        self,
        mols: Sequence[Union[str, dm.Mol]],
        n_jobs: Optional[int] = -1,
        progress: bool = False,
        progress_leave: bool = False,
        scheduler: str = "auto",
        batch_size: Optional[int] = None,
        keep_details: bool = False,
    ) -> pd.DataFrame:
        """Run alert evaluation on this list of molecule and return the full dataframe

        Args:
            mols: list of input molecule object.
            n_jobs: number of jobs to run in parallel.
            progress: whether to show progress or not.
            progress_leave: whether to leave the progress bar or not.
            scheduler: which scheduler to use. If "auto", will use "processes" if `len(mols) > 500` else "threads".
            batch_size: batch size to use for parallelization.
            keep_details: whether to keep the details of the evaluation or not.
        """

        if scheduler == "auto":
            if len(mols) > 500:
                scheduler = "processes"  # pragma: no cover
            else:
                scheduler = "threads"
        if batch_size:
            results = dm.parallelized_with_batches(
                lambda batch: [
                    functools.partial(self._evaluate, keep_details=keep_details)(mol) for mol in batch
                ],
                mols,
                progress=progress,
                n_jobs=n_jobs,
                scheduler=scheduler,
                batch_size=batch_size,
                tqdm_kwargs=dict(
                    desc="Common alerts filtering",
                    leave=progress_leave,
                ),
            )
        else:
            results = dm.parallelized(
                functools.partial(self._evaluate, keep_details=keep_details),
                mols,
                progress=progress,
                n_jobs=n_jobs,
                scheduler=scheduler,
                tqdm_kwargs=dict(
                    desc="Common alerts filtering",
                    leave=progress_leave,
                ),
            )
        results = pd.DataFrame(results)

        return results



import importlib.resources

import pandas as pd

from medchem.utils.loader import get_data_path
from medchem.utils.loader import get_grammar


def test_get_data_path():
    df = pd.read_csv(get_data_path("medchem_rule_list.csv"))
    assert df.columns.tolist() == ["name", "rules", "description"]


def test_get_grammar_default():
    with open(get_grammar()) as f:
        grammar_text = f.read()

    assert "bool_term (OR_OP bool_term)" in grammar_text


def test_get_grammar_as_string():
    grammar_text = get_grammar(as_string=True)
    assert "bool_term (OR_OP bool_term)" in grammar_text


def test_get_grammar_custom():
    path = importlib.resources.files("medchem.data").joinpath("grammar.lark")
    path = str(path)

    with open(get_grammar(path)) as f:
        grammar_text = f.read()

    assert "bool_term (OR_OP bool_term)" in grammar_text



from typing import Optional
from typing import cast
from typing import Union

import os
import io
import sys
import functools

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources

import fsspec


@functools.lru_cache(maxsize=10)
def get_data_path(filename: str, module: str = "medchem.data"):
    """Return the filepath of an internal data file."""

    if sys.version_info < (3, 9, 0):
        with importlib_resources.path(module, filename) as p:
            path = p
    else:
        path = importlib_resources.files(module).joinpath(filename)
    return str(path)


def get_grammar(
    grammar: Optional[Union[os.PathLike, str]] = None,
    as_string: bool = False,
):
    """Return the default lark grammar file for the medchem query system

    Args:
        grammar: The path to the grammar file. If None, the default medchem grammar file is used.
        as_string: If True, return the grammar as a string. Defaults to False.
    """
    if grammar is None:
        _grammar = get_data_path("grammar.lark")
        _grammar = str(_grammar)
    else:
        _grammar = str(grammar)

    if as_string:
        with fsspec.open(_grammar, "r") as f:
            f = cast(io.TextIOBase, f)
            _grammar = f.read()

    return _grammar



from typer.testing import CliRunner

import datamol as dm

from medchem.cli import app

runner = CliRunner()


def test_cli():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0


def test_common_alerts(tmp_path):
    input_path = str(tmp_path / "input.csv")
    output_path = str(tmp_path / "output.csv")

    # Load a dataset
    data = dm.data.solubility()
    data = data.sample(50, random_state=20)

    data.to_csv(input_path, index=False)

    result = runner.invoke(
        app,
        [
            "common-alerts",
            str(input_path),
            str(output_path),
            "--smiles-column",
            "smiles",
            "--keep-details",
        ],
    )
    assert result.exit_code == 0

    # Check the output
    results = dm.io.open_df(output_path)
    assert set(results.columns) == {"smiles", "pass_filter", "status", "reasons", "details"}
    assert results.shape[0] == 50


def test_nibr_filters(tmp_path):
    input_path = str(tmp_path / "input.csv")
    output_path = str(tmp_path / "output.csv")

    # Load a dataset
    data = dm.data.solubility()
    data = data.sample(50, random_state=20)

    data.to_csv(input_path, index=False)

    result = runner.invoke(
        app,
        [
            "nibr-filters",
            str(input_path),
            str(output_path),
            "--smiles-column",
            "smiles",
            "--keep-details",
        ],
    )
    assert result.exit_code == 0

    # Check the output
    results = dm.io.open_df(output_path)
    assert set(results.columns) == {
        "smiles",
        "pass_filter",
        "reasons",
        "severity",
        "status",
        "n_covalent_motif",
        "special_mol",
        "details",
    }
    assert results.shape[0] == 50



from typing import Sequence
from typing import Optional
from typing import Callable
from typing import Union

import itertools

import numpy as np
import datamol as dm

from rdkit.Chem import rdMolDescriptors  # type: ignore

from ..utils.graph import score_symmetry


def _generic_filter(
    mols: Sequence[Union[str, dm.Mol]],
    rejection_fn: Callable,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Apply any generic filters to a molecule

    ??? tip "True is good"
        Returning `True` means the molecule does not fail the rejection function.

    Args:
        mols: list of input molecules
        rejection_fn: function that apply rejection rules on a single molecule. Return True is the molecule is rejected.
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to keep or leave the progress bar after completion
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """
    if isinstance(mols[0], str):
        mols = dm.parallelized(
            dm.to_mol,
            mols,
            n_jobs=n_jobs,
            progress=progress,
            tqdm_kwargs=dict(desc="To mol", leave=progress_leave),
        )

    toxic = dm.parallelized(
        rejection_fn,
        mols,
        n_jobs=n_jobs,
        scheduler=scheduler,
        progress=progress,
        tqdm_kwargs=dict(desc="Match", leave=progress_leave),
    )

    toxic = np.array(toxic)
    filtered_idx = np.where(~toxic)[0]

    if return_idx:
        return filtered_idx

    return np.bitwise_not(toxic)


def macrocycle_filter(
    mols: Sequence[Union[str, dm.Mol]],
    max_cycle_size: int = 10,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Find valid molecules that do not infringe the strict maximum cycle size.


    ??? tip "True is good"
        Returning `True` means the molecule does not have rings larger than `max_cycle_size`

    Args:
        mols: list of input molecules
        max_cycle_size: strict maximum macrocycle size
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.

    """

    def reject_fn(mol):
        if mol is None:
            return True
        rinfo = mol.GetRingInfo().AtomRings()
        return any(len(r) >= max_cycle_size for r in rinfo)

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def atom_list_filter(
    mols: Sequence[Union[str, dm.Mol]],
    unwanted_atom_list: Optional[Sequence] = None,
    wanted_atom_list: Optional[Sequence] = None,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Find molecules without any atom from a set of unwanted atomic symbols
    and with all atoms in the set of wanted atom list.

    ??? tip "True is good"
        Returning `True` means the molecule only has desirable atom types

    Args:
        mols: list of input molecules
        unwanted_atom_list: list of undesirable atomic symbol
        wanted_atom_list: list of desirable atomic symbol
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel jobs to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """
    unwanted_atom_list = unwanted_atom_list or []
    wanted_atom_list = wanted_atom_list or []

    def reject_fn(mol):
        if mol is None:
            return True
        for atom in mol.GetAtoms():
            cur_symb = atom.GetSymbol()
            if cur_symb in unwanted_atom_list or (
                len(wanted_atom_list) > 0 and cur_symb not in wanted_atom_list
            ):
                return True
        return False

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def ring_infraction_filter(
    mols: Sequence[Union[str, dm.Mol]],
    hetcycle_min_size: int = 4,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """
    Find molecules that have a ring infraction filter. This filter focuses on checking for rings that are too small to have an heteroatom.

    ??? tip "True is good"
        Returning `True` means the molecule does not infringe the ring infraction filter.


    Args:
        mols: list of input molecules
        hetcycle_min_size: Minimum ring size before more than 1 hetero atom or any non single bond is allowed.
            This is a *strict threshold (>)*
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """

    ring_allene = dm.from_smarts("[R]=[R]=[R]")
    double_bond_small_ring = dm.from_smarts("[r3,r4]=[r3,r4]")

    def reject_fn(mol):
        if mol is None:
            return True
        rejected = mol.HasSubstructMatch(ring_allene) or mol.HasSubstructMatch(double_bond_small_ring)
        if rejected:
            return True
        rinfo = mol.GetRingInfo()
        bond_rings = rinfo.BondRings()
        for r in bond_rings:
            r_bonds = [mol.GetBondWithIdx(b) for b in r]
            r_bond_types = [b.GetBondType() for b in r_bonds]
            r_atom_content = set(
                itertools.chain(
                    *[(b.GetBeginAtom().GetSymbol(), b.GetEndAtom().GetSymbol()) for b in r_bonds]
                )
            )
            n_ring_heteroatoms = sum([at not in ["C", "H"] for at in r_atom_content])
            if len(r) <= hetcycle_min_size and (
                n_ring_heteroatoms > 1 or any(btype != dm.SINGLE_BOND for btype in r_bond_types)
            ):
                # too many heteroatoms in low ring size OR
                # alkene, allen or aromatic in small rings
                return True
        return False

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def num_atom_filter(
    mols: Sequence[Union[str, dm.Mol]],
    min_atoms: Optional[int] = None,
    max_atoms: Optional[int] = None,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """
    Find molecules that match the number of atom range constraints

    ??? tip "True is good"
        Returning `True` means the molecule does not infringe the number of atom filter.


    Args:
        mols: list of input molecules
        min_atoms: strict minimum number of atoms (atoms > min_atoms)
        max_atoms: strict maximum number of atoms (atoms < max_atoms)
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """

    def reject_fn(mol):
        if mol is None:
            return True
        num_atoms = mol.GetNumAtoms()
        return (min_atoms is not None and num_atoms <= min_atoms) or (
            max_atoms is not None and num_atoms >= max_atoms
        )

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def num_stereo_center_filter(
    mols: Sequence[Union[str, dm.Mol]],
    max_stereo_centers: int = 4,
    max_undefined_stereo_centers: int = 2,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """
    Find molecules that match the number of stereo center constraints. In general, molecules with
    too many undefined stereo centers are not desirable. This filter is useful for generated molecules.


    ??? tip "True is good"
        Returning `True` means the molecule does not have issues with stereo centers.

    Args:
        mols: list of input molecules
        max_stereo_centers: strict maximum number of stereo centers (<). Default is 4
        max_undefined_stereo_centers: strict maximum number of undefined stereo centers (<). Default is 2
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """

    def reject_fn(mol):
        if mol is None:
            return True
        return (dm.descriptors.n_stereo_centers(mol) >= max_stereo_centers) or (
            rdMolDescriptors.CalcNumUnspecifiedAtomStereoCenters(mol) >= max_undefined_stereo_centers
        )

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def halogenicity_filter(
    mols: Sequence[Union[str, dm.Mol]],
    thresh_F: int = 6,
    thresh_Br: int = 3,
    thresh_Cl: int = 3,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Find molecules that do not exceed halogen count threshold. This filter is useful for removing halogen biases
    in generated or enumerated chemical space during goal-directed optimization.

    - 6 for fluorine
    - 3 for bromine
    - 3 for chlorine

    ??? tip "True is good"
        Returning `True` means the molecule does not have too many halogen atoms.

    Args:
        mols: list of input molecules
        thresh_F: maximum number of fluorine
        thresh_Br: maximum number of bromine
        thresh_Cl: maximum number of chlorine
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """
    fluorine_smarts = dm.from_smarts("[F]")
    bromine_smarts = dm.from_smarts("[Br]")
    chlorine_smarts = dm.from_smarts("[Cl]")

    def reject_fn(mol):
        if mol is None:
            return True
        fluorine_saturation = len(mol.GetSubstructMatches(fluorine_smarts, uniquify=True)) > thresh_F
        bromine_saturation = len(mol.GetSubstructMatches(bromine_smarts, uniquify=True)) > thresh_Br
        chlorine_saturation = len(mol.GetSubstructMatches(chlorine_smarts, uniquify=True)) > thresh_Cl
        return fluorine_saturation or bromine_saturation or chlorine_saturation

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )


def symmetry_filter(
    mols: Sequence[Union[str, dm.Mol]],
    symmetry_threshold: float = 0.8,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Find molecules that are not symmetrical, given a symmetry threshold.
    This filter was designed to offset the symmetry issue in molecular design,
    where some models tend to generate highly symmetrical molecules due to substructure bias.

    ??? tip "True is good"
        Returning `True` means the molecule is not too symmetrical

    Args:
        mols: list of input molecules
        symmetry_threshold: threshold to consider a molecule highly symmetrical
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """

    def reject_fn(mol):
        return score_symmetry(mol) > symmetry_threshold

    return _generic_filter(
        mols,
        reject_fn,
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
    )



from importlib.metadata import version
from importlib.metadata import PackageNotFoundError

try:
    __version__ = version("medchem")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"



from typing import cast

import datamol as dm

from medchem.utils.smarts import SMARTSUtils


def test_standard_attachment():
    smiles = [
        "C1CC(C)C(O)CC1[1*]",  # ortho
        "c1c(C)cc(O)cc1[*:2]",  # meta + aromatic
        "c1(C)ccc(O)cc1[*]",  # para + aromat
    ]

    for sm in smiles:
        new_sm = SMARTSUtils.standardize_attachment(sm, "[*:1]")
        assert "[*:1]" in new_sm


def test_ortho_meta_para():
    smiles = [
        "C1CC(C)C(O)CC1",  # ortho
        "c1c(C)cc(O)cc1",  # meta + aromatic
        "c1(C)ccc(O)cc1",  # para + aromat
    ]
    mols = [dm.to_mol(x) for x in smiles]
    mols = [cast(dm.Mol, x) for x in mols]

    sm1 = "[#6;!R]"
    sm2 = "[#8]"
    ortho_query = dm.from_smarts(SMARTSUtils.ortho(sm1, sm2))
    ortho_aro_query = dm.from_smarts(SMARTSUtils.ortho(sm1, sm2, aromatic_only=True))
    meta_query = dm.from_smarts(SMARTSUtils.meta(sm1, sm2))
    para_query = dm.from_smarts(SMARTSUtils.para(sm1, sm2))
    expected_output = [True, True, True, False, False]
    output = [
        mols[0].HasSubstructMatch(ortho_query),  # True
        mols[1].HasSubstructMatch(meta_query),  # True
        mols[2].HasSubstructMatch(para_query),  # True
        mols[0].HasSubstructMatch(para_query),  # False
        mols[0].HasSubstructMatch(ortho_aro_query),  # False
    ]

    assert output == expected_output


def test_long_chain():
    smiles = [
        "CCCCCCCC",  # octane
        "CCCC=CCCN",  #  hept-3-en-1-amine
        "CCOCCC",  #  1-ethoxypropane
        "CC(C)(C)CN",  #  2,2-dimethylpropan-1-amine
        "CCC(C)(C)CN",  # 2,2-dimethylbutan-1-amine
        "C1CCCC1CC",  # ethylcyclopentane
    ]
    mols = [dm.to_mol(x) for x in smiles]
    mols = [cast(dm.Mol, x) for x in mols]

    chain1 = dm.from_smarts(
        SMARTSUtils.aliphatic_chain(
            min_size=5,
            unsaturated_bondtype=None,
        ),
    )
    chain2 = dm.from_smarts(
        SMARTSUtils.aliphatic_chain(
            min_size=6,
            unbranched=True,
            unsaturated_bondtype=dm.SINGLE_BOND,
        )
    )
    chain3 = dm.from_smarts(
        SMARTSUtils.aliphatic_chain(
            min_size=6,
            unbranched=True,
            unsaturated_bondtype=dm.DOUBLE_BOND,
            allow_hetero_atoms=False,
        )
    )
    chain1_out = [m.HasSubstructMatch(chain1) for m in mols]
    chain2_out = [m.HasSubstructMatch(chain2) for m in mols]
    chain3_out = [m.HasSubstructMatch(chain3) for m in mols]

    assert chain1_out == [True, True, True, False, True, False]
    assert chain2_out == [True, False, True, False, False, False]
    assert chain3_out == [True, True, False, False, False, False]


def test_atom_in_env():
    mol = dm.to_mol("c1[c:1](OC)cc(F)cc1")
    mol = cast(dm.Mol, mol)

    # we are trying to match the carbon with atom map 1
    # these queries should be equivalent
    atom_id = tuple([atom.GetIdx() for atom in mol.GetAtoms() if atom.GetAtomMapNum() == 1])
    query1 = SMARTSUtils.atom_in_env("[#6;r6]", "[*][OD2][C&D1]", "[c]aa[F]", union=False)
    query2 = SMARTSUtils.atom_in_env("[#6;r6][OD2][C&D1]", "[c]aa[F]", union=False)

    res1 = mol.GetSubstructMatch(dm.from_smarts(query1))
    res2 = mol.GetSubstructMatch(dm.from_smarts(query2))

    assert res1 == res2
    assert res1 == atom_id



import datamol as dm
import pandas as pd

from rdkit.Chem import rdfiltercatalog  # type: ignore

from medchem.groups import ChemicalGroup
from medchem.groups import list_default_chemical_groups
from medchem.groups import list_functional_group_names
from medchem.groups import get_functional_group_map

from medchem.utils.loader import get_data_path


def test_list_default_chemical_groups():
    assert len(list_default_chemical_groups()) > 0
    assert all([isinstance(ll, str) for ll in list_default_chemical_groups()])

    with_hierarchy = list_default_chemical_groups(hierarchy=True)
    assert len(with_hierarchy) > 0
    assert all([isinstance(ll, str) for ll in with_hierarchy])
    assert all(["." in ll for ll in with_hierarchy])


def test_list_functional_group_names():
    group_names = list_functional_group_names()

    assert len(group_names) > 0
    assert all([isinstance(ll, str) for ll in group_names])


def test_get_functional_group_map():
    group_map = get_functional_group_map()
    assert len(group_map) > 0
    assert isinstance(group_map, dict)


def test_chemical_group_filter():
    group = ChemicalGroup(groups="rings_in_drugs")
    assert len(group.data) == 92
    group.filter(["azole", "3,4-dihydroquino-2(1H)-one"], fuzzy=False)
    assert len(group.data) == 1

    group = ChemicalGroup(groups="rings_in_drugs")

    assert len(group.data) == 92
    group.filter(["azole", "3,4-dihydroquino-2(1H)-one"], fuzzy=True)
    assert len(group.data) == 12

    group = ChemicalGroup(groups="rings_in_drugs")

    assert len(group.data) == 92
    group.filter([])
    assert len(group.data) == 92


def test_chemical_group_attributes():
    group = ChemicalGroup(groups="rings_in_drugs")
    assert isinstance(group.name, list)
    assert isinstance(group.smiles, list)
    assert isinstance(group.mols, list)
    assert isinstance(group.smarts, list)
    assert isinstance(group.mol_smarts, list)
    assert isinstance(group.dataframe, pd.DataFrame)


def test_chemical_group_list():
    group = ChemicalGroup(groups="rings_in_drugs")
    assert isinstance(group.list_groups(), list)
    assert isinstance(group.list_hierarchy_groups(), list)


def test_chemical_group_catalogs():
    group = ChemicalGroup(groups="rings_in_drugs")
    assert isinstance(group.get_catalog(), rdfiltercatalog.FilterCatalog)


def test_chemical_group():
    c_group = ChemicalGroup(groups="rings_in_drugs")
    rings_numbers = 92
    assert len(c_group) == rings_numbers
    assert len(c_group.smarts) == rings_numbers

    out_false = c_group.has_match("CCCCCCCCCC")
    assert out_false is False

    out_true = c_group.has_match("C1CCCCC1")
    assert out_true is True


def test_chemical_group_query():
    c_group = ChemicalGroup(groups="rings_in_drugs")
    mol = dm.to_mol("CCS(=O)(=O)N1CC(C1)(CC#N)N2C=C(C=N2)C3=C4C=CNC4=NC=N3")

    # we have two rings of interest for smarts
    out = c_group.get_matches(mol, use_smiles=False)

    assert out is not None
    assert out.shape[0] == 2
    assert set(out.group.unique()) == {"rings_in_drugs"}
    assert set(out.name.unique()) == {"diazine", "1H-pyrrole"}

    # however, if we use smiles, we would have 3
    # this is a bug that needs to be fixed
    out_smiles = c_group.get_matches(mol, use_smiles=True)
    assert out_smiles is not None
    assert out_smiles.shape[0] == 3
    assert set(out_smiles.group.unique()) == {"rings_in_drugs"}
    assert set(out_smiles.name.unique()) == {"diazine", "1H-pyrrole", "1H-pyrazole"}


def test_external_bank():
    c_group = ChemicalGroup(groups_db=get_data_path("smarts_bank.csv"))
    mol = dm.to_mol("CCS(=O)(=O)N1CC(C1)(CC#N)N2C=C(C=N2)C3=C4C=CNC4=NC=N3")
    out = c_group.get_matches(mol, use_smiles=False)
    expected_match = set(["HBA", "HBD", "Hydrogen", "SP3 Nitrogen", "SP2 Carbon"])

    assert out is not None
    assert expected_match.issubset(set(out["name"].values)) is True



from typing import Sequence
from typing import Optional
from typing import Union

import functools

import pandas as pd
import datamol as dm

from medchem.catalogs import NamedCatalogs


class NIBRFilters:
    """Filtering class for building a screening deck following the novartis filtering process
    published in https://dx.doi.org/10.1021/acs.jmedchem.0c01332.

    The output of the filter are explained below:
    - **status**: one of `["exclude", "flag", "annotations", "ok"]` (ordered by quality).
        Generally, you can keep anything without the "exclude" label, as long as you also apply
        a maximum severity score for compounds that collects too many flags.
    - **n_covalent_motif**: number of potentially covalent motifs contained in the compound
    - **severity**: how severe are the issues with the molecules:
        - `0`: compound has no flags, might have annotations;
        - `1-9`:  number of flags the compound raises;
        - `>= 10`:  default exclusion criterion used in the paper
    - **special_mol**: whether the compound/parts of the compound belongs to a special class of molecules
        (e.g peptides, glycosides, fatty acid). In that case, you should review the rejection reasons.
    - **pass_filter**: whether the compound passed the filter or not.
    - **details**: optional additional details of the evaluation, including matching patterns if `keep_details` is True.
    """

    def __init__(self):
        self.catalog = NamedCatalogs.nibr()

    def _evaluate(
        self,
        mol: Union[str, dm.Mol],
        keep_details: bool = False,
    ):
        """Evaluate one molecule."""

        if isinstance(mol, str):
            with dm.without_rdkit_log():
                _mol = dm.to_mol(mol)
        else:
            _mol = mol

        if _mol is None:
            return pd.Series(
                {
                    "mol": _mol,
                    "pass_filter": False,
                    "reasons": "invalid",
                    "severity": 0,
                    "status": "exclude",
                    "n_covalent_motif": 0,
                    "special_mol": 0,
                }
            )

        # Get the matches
        entries = self.catalog.GetMatches(_mol)

        if len(entries) == 0:
            return pd.Series(
                {
                    "mol": _mol,
                    "pass_filter": True,
                    "reasons": None,
                    "severity": 0,
                    "status": "ok",
                    "n_covalent_motif": 0,
                    "special_mol": 0,
                }
            )

        result = pd.Series()
        result["mol"] = _mol

        # Iterate over all the matchign entries
        result_entries = []
        for entry in entries:
            pname = entry.GetDescription()
            _, name, severity, n_covalent_motif, special_mol = pname.split("||")

            result_entry = {}
            result_entry["name"] = name
            result_entry["severity"] = int(severity)
            result_entry["n_covalent_motif"] = int(n_covalent_motif)
            result_entry["special_mol"] = bool(special_mol)
            result_entries.append(result_entry)

        result_entries = pd.DataFrame(result_entries)

        # Now build a flat result from the detailed results per entries

        result["reasons"] = "; ".join(result_entries["name"].tolist())

        # severity of 2 means EXCLUDE
        if 2 in result_entries["severity"].values:
            result["severity"] = 10
            result["status"] = "exclude"
        else:
            result["severity"] = result_entries["severity"].sum()

            if 1 in result_entries["severity"].values:
                result["status"] = "flag"
            elif 0 in result_entries["severity"].values:
                result["status"] = "annotations"

        # get number of covalent flags and special molecule flags
        result["n_covalent_motif"] = result_entries["n_covalent_motif"].sum()
        result["special_mol"] = result_entries["special_mol"].sum()

        if result["status"] == "exclude":
            result["pass_filter"] = False
        else:
            result["pass_filter"] = True

        if keep_details:
            result["details"] = result_entries.to_dict(orient="records")

        return result

    def __call__(
        self,
        mols: Sequence[Union[str, dm.Mol]],
        n_jobs: Optional[int] = -1,
        progress: bool = False,
        progress_leave: bool = False,
        scheduler: str = "threads",
        keep_details: bool = False,
    ):
        """Run alert evaluation on this list of molecule and return the full dataframe

        Args:
            mols: list of input molecule object.
            n_jobs: number of jobs to run in parallel.
            progress: whether to show progress or not.
            progress_leave: whether to leave the progress bar or not.
            scheduler: which scheduler to use. The `processes` scheduler works but is very
                inefficient due to RDKit Catalog serialization which tends to be very slow.
            keep_details: whether to keep the details of the evaluation or not.
        """

        results = dm.parallelized(
            functools.partial(self._evaluate, keep_details=keep_details),
            mols,
            progress=progress,
            n_jobs=n_jobs,
            scheduler=scheduler,
            tqdm_kwargs=dict(
                desc="NIBR filtering",
                leave=progress_leave,
            ),
        )
        results = pd.DataFrame(results)

        return results



from typing import cast
from typing import Union
from typing import List
from typing import Any

import datamol as dm
import numpy as np
import networkx as nx

from rdkit.Chem import SaltRemover

DEFAULT_NODE_ATTR = ["atomic_num", "degree", "ring_atom", "is_aromatic"]
DEFAULT_EDGE_ATTR = ["bond_type"]


def automorphism(
    mol: Union[str, dm.Mol],
    standardize: bool = True,
    node_attrs: List[str] = DEFAULT_NODE_ATTR,
    edge_attrs: List[str] = DEFAULT_EDGE_ATTR,
):
    """Compute automorphism in a molecular graph

    Args:
        mol: input molecular graph
        standardize: whether to standardize the compound or not
        node_attrs: list of categorical atom attributes/properties to consider for node matching
        edge_attrs: list of categorical bond attributes/properties to consider for edge matching
    """

    # there is no need to kekulize the molecule, in order to preserve
    # aromaticity instead of switching to double bonds
    with dm.without_rdkit_log():
        mol = dm.to_mol(mol, kekulize=False)
        # important to remove hydrogen to reduce automorphism
        # similarly there is no point in considering stereochemistry
        mol = dm.remove_hs(mol)

        if standardize:
            mol = dm.standardize_mol(mol, disconnect_metals=True, uncharge=True)
            # we also remove salts
            remover = SaltRemover.SaltRemover()
            mol = remover.StripMol(mol, dontRemoveEverything=True)

    # the graph is undirected by default
    graph = dm.graph.to_graph(mol)
    graph_copy = graph.copy()

    node_match, edge_match = None, None
    if node_attrs:
        node_match = nx.algorithms.isomorphism.categorical_node_match(node_attrs, [0] * len(node_attrs))
    if edge_attrs:
        edge_match = nx.algorithms.isomorphism.categorical_edge_match(edge_attrs, [0] * len(node_attrs))

    graph_matcher = nx.algorithms.isomorphism.GraphMatcher(
        graph, graph_copy, node_match=node_match, edge_match=edge_match
    )
    matches = list(graph_matcher.match())
    return dict(graph=graph, matches=matches, mol=mol)


def score_symmetry(
    mol: Union[dm.Mol, str], exclude_self_mapped_edged: bool = False, **automorphism_kwargs: Any
):
    """Provide a symmetry score for a given input molecule

    !!! note
        This is an heuristic and our definition of symmetry is pretty loose.
        We define symmetry according to any (set of) plans dividing the molecule into two very similar subgraph.
        We include both edge and vertex transitivity. For example the star-molecular graph
        (e.g neopentane) is symmetrical here, although it's not vertex-transitive.

    Args:
        mol: inputs molecules
        exclude_self_mapped_edged: Whether to exclude edges that matches to themselves in automorphism.
        automorphism_kwargs: keyword for determining automorphism
    """

    def _is_self_mapped(mapping):
        return all(node1 == node2 for node1, node2 in mapping.items())

    mol_automorphism = automorphism(mol, **automorphism_kwargs)

    # We check and filter for identity mapping x->x for all nodes
    graph = mol_automorphism["graph"]
    graphmol = mol_automorphism["mol"]
    matches = mol_automorphism["matches"]
    matches = [x for x in matches if not _is_self_mapped(x)]

    # Always make the type explicit
    graphmol = cast(dm.Mol, graphmol)
    graph = cast(nx.Graph, graph)
    matches = cast(List[dict], matches)

    # Step 1: We set the symmetry score to 0 when there is not automorphism
    if len(matches) == 0:
        return 0

    # Step 2: let's find the coverage of edge satisfying the symmetric graph notion
    # https://en.wikipedia.org/wiki/Symmetric_graph

    arc_transitivity = []
    for bond in graphmol.GetBonds():
        edge = tuple(sorted([bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()]))
        is_covered = False
        for mapping in matches:
            iso_edge = (mapping.get(edge[0]), mapping.get(edge[1]))
            if graph.has_edge(*iso_edge) and (iso_edge != edge):
                is_covered = True
                break
            elif set(iso_edge) == set(edge) and not exclude_self_mapped_edged:
                # we only accept those when at least some neighbors of the nodse in the edge map to each other
                neighbors1 = set(graph.neighbors(edge[0]))
                neighbors2 = set(graph.neighbors(edge[1]))
                if any((mapping.get(x) in neighbors1 - {x}) for x in neighbors1) and any(
                    (mapping.get(x) in neighbors2 - {x}) for x in neighbors2
                ):
                    is_covered = True
                    break
        arc_transitivity.append(is_covered)
    return np.mean(arc_transitivity)



from typing import Union
from typing import Optional
from typing import Any

import datamol as dm

from ._utils import in_range
from ._utils import n_fused_aromatic_rings
from ._utils import n_heavy_metals
from ._utils import has_spider_chains
from ._utils import fraction_atom_in_scaff


def rule_of_five(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_lipinski_hbd: Optional[float] = None,
    n_lipinski_hba: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """Compute the Lipinski's rule-of-5 for a molecule. Also known as Pfizer's rule of five or RO5.
    This rule is a rule of thumb to evaluate the druglikeness of a chemical compounds.

    It computes: `MW <= 500 & logP <= 5  & HBD <= 5 & HBA <= 10`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_lipinski_hbd: precomputed number of HBD.
        n_lipinski_hba: precomputed number of HBA.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro5: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_lipinski_hbd = n_lipinski_hbd if n_lipinski_hbd is not None else dm.descriptors.n_lipinski_hbd(mol)
    n_lipinski_hba = n_lipinski_hba if n_lipinski_hba is not None else dm.descriptors.n_lipinski_hba(mol)
    return mw <= 500 and clogp <= 5 and n_lipinski_hbd <= 5 and n_lipinski_hba <= 10


def rule_of_five_beyond(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hbd: Optional[float] = None,
    n_hba: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """Compute the Beyond rule-of-5 rule for a molecule. This rule illustrates the potential of compounds far beyond rule-of-5 space to
    modulate novel and difficult target classes that have large, flat, and groove-shaped binding sites and has been described in:

    [Doak, Bradley C., et al. (2015) How Beyond Rule of 5 Drugs and Clinical Candidates Bind to Their Targets.](https://pubmed.ncbi.nlm.nih.gov/26457449/)

    It computes: `MW <= 1000 & logP in [-2, 10] & HBD <= 6 & HBA <= 15 & TPSA <=250 & ROTBONDS <= 20`

    !!! note
        This is a very permissive rule and is likely to not be a good predictor for druglikeness as known for small molecules.

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hbd: precomputed number of HBD.
        n_hba: precomputed number of HBA.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro5: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    return (
        mw <= 1000
        and in_range(clogp, -2, 10)
        and n_hbd <= 6
        and n_hba <= 15
        and tpsa <= 250
        and n_rotatable_bonds <= 20
    )


def rule_of_zinc(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    charge: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """Compute the Zinc rule for a molecule. This rule is a rule of thumb to evaluate the druglikeness of a chemical compounds, based on:

    [Irwin & Schoichet (2005) ZINC - A Free Database of Commercially Available Compounds for Virtual Screening](https://pubmed.ncbi.nlm.nih.gov/15667143/)

     Also see: https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html

     It computes: `MW in [60, 600] & logP < in [-4, 6] & HBD <= 6 & HBA <= 11 & TPSA <=150 & ROTBONDS <= 12 & RIGBONDS <= 50 & N_RINGS <= 7 & MAX_SIZE_RING <= 12 & N_CARBONS >=3 & HC_RATIO <= 2.0 & CHARGE in [-4, 4]`

     Args:
         mol: input molecule
         mw: precomputed molecular weight.
         clogp: precomputed cLogP.
         n_hba: precomputed number of HBA.
         n_hbd: precomputed number of HBD.
         tpsa: precomputed TPSA.
         n_rotatable_bonds: precomputed number of rotatable bonds.
         n_rings: precomputed number of rings in the molecules.
         charge: precomputed charge.
         **kwargs: Allow extra arguments for descriptors pre-computation.

    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)
    n_rigid_bonds = dm.descriptors.n_rigid_bonds(mol)
    ring_system = dm.compute_ring_system(mol, include_spiro=False)
    max_size_ring = 0 if len(ring_system) == 0 else max([len(x) for x in ring_system])
    n_carbons = len([at for at in mol.GetAtoms() if at.GetSymbol() == "C"])
    if n_carbons == 0:
        het_carb_ratio = float("inf")
    else:
        het_carb_ratio = dm.descriptors.n_hetero_atoms(mol) / n_carbons
    charge = charge if charge is not None else dm.descriptors.formal_charge(mol)
    return (
        in_range(mw, 60, 600)
        and in_range(clogp, -4, 6)
        and n_hbd <= 6
        and n_hba <= 11
        and tpsa <= 150
        and n_rotatable_bonds <= 12
        and n_rigid_bonds <= 50
        and n_rings <= 7
        and max_size_ring <= 12
        and n_carbons >= 3
        and in_range(het_carb_ratio, 0, 2.0)
        and in_range(charge, -4, 4)
    )


def rule_of_leadlike_soft(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    n_hetero_atoms: Optional[int] = None,
    charge: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the Lead-Like Soft rule available in [FAF-Drugs4](https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html).

    It computes:
    ```
    MW in [150, 400] & logP < in [-3, 4] & HBD <= 4 & HBA <= 7 & TPSA <=160 & ROTBONDS <= 9 &
    RIGBONDS <= 30 & N_RINGS <= 4 & MAX_SIZE_RING <= 18 & N_CARBONS in [3, 35] &  N_HETEROATOMS in [1, 15] &
    HC_RATIO in [0.1, 1.1] & CHARGE in [-4, 4] & N_ATOM_CHARGE <= 4 & N_STEREO_CENTER <= 2
    ```

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        n_rings: precomputed number of rings in the molecules.
        n_hetero_atoms: precomputed number of heteroatoms.
        charge: precomputed charge.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)
    n_hetero_atoms = n_hetero_atoms if n_hetero_atoms is not None else dm.descriptors.n_hetero_atoms(mol)
    charge = charge if charge is not None else dm.descriptors.formal_charge(mol)
    num_charged_atom = dm.descriptors.n_charged_atoms(mol)
    n_stereo_center = dm.descriptors.n_stereo_centers(mol)
    n_rigid_bonds = dm.descriptors.n_rigid_bonds(mol)
    ring_system = dm.compute_ring_system(mol, include_spiro=False)
    max_size_ring = 0 if len(ring_system) == 0 else max([len(x) for x in ring_system])
    n_carbons = len([at for at in mol.GetAtoms() if at.GetSymbol() == "C"])
    if n_carbons == 0:
        het_carb_ratio = float("inf")
    else:
        het_carb_ratio = n_hetero_atoms / n_carbons
    return (
        in_range(mw, 150, 400)
        and in_range(clogp, -3, 4)
        and n_hbd <= 4
        and n_hba <= 7
        and tpsa <= 160
        and n_rotatable_bonds <= 9
        and n_rigid_bonds <= 30
        and n_rings <= 4
        and max_size_ring <= 18
        and in_range(n_carbons, 3, 35)
        and in_range(n_hetero_atoms, 1, 15)
        and in_range(het_carb_ratio, 0.1, 1.1)
        and in_range(charge, -4, 4)
        and num_charged_atom <= 4
        and n_stereo_center <= 2
    )


def rule_of_druglike_soft(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    n_hetero_atoms: Optional[int] = None,
    charge: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the DrugLike Soft rule available in [FAF-Drugs4](https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html).

    It computes:
    ```
    MW in [100, 600] & logP < in [-3, 6] & HBD <= 7 & HBA <= 12 & TPSA <=180 & ROTBONDS <= 11 &
    RIGBONDS <= 30 & N_RINGS <= 6 & MAX_SIZE_RING <= 18 & N_CARBONS in [3, 35] &  N_HETEROATOMS in [1, 15] &
    HC_RATIO in [0.1, 1.1] & CHARGE in [-4, 4] & N_ATOM_CHARGE <= 4
    ```

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        n_rings: precomputed number of rings in the molecules.
        n_hetero_atoms: precomputed number of heteroatoms.
        charge: precomputed charge.
        **kwargs: Allow extra arguments for descriptors pre-computation.
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)
    n_hetero_atoms = n_hetero_atoms if n_hetero_atoms is not None else dm.descriptors.n_hetero_atoms(mol)
    charge = charge if charge is not None else dm.descriptors.formal_charge(mol)
    num_charged_atom = dm.descriptors.n_charged_atoms(mol)
    n_rigid_bonds = dm.descriptors.n_rigid_bonds(mol)
    ring_system = dm.compute_ring_system(mol, include_spiro=False)
    max_size_ring = 0 if len(ring_system) == 0 else max([len(x) for x in ring_system])
    n_carbons = len([at for at in mol.GetAtoms() if at.GetSymbol() == "C"])
    n_hydrogens = sum([at.GetTotalNumHs() for at in mol.GetAtoms()])

    if n_carbons == 0:
        het_carb_ratio = float("inf")
    else:
        het_carb_ratio = n_hydrogens / n_carbons

    return (
        in_range(mw, 100, 600)
        and in_range(clogp, -3, 6)
        and n_hbd <= 7
        and n_hba <= 12
        and tpsa <= 180
        and n_rotatable_bonds <= 11
        and n_rigid_bonds <= 30
        and n_rings <= 6
        and max_size_ring <= 18
        and in_range(n_carbons, 3, 35)
        and in_range(n_hetero_atoms, 1, 15)
        and in_range(het_carb_ratio, 0.1, 1.1)
        and in_range(charge, -4, 4)
        and num_charged_atom <= 4
    )


def rule_of_four(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_rings: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """Compute the rule-of-4 for a molecule. The rule-of-4 define a rule of thumb for PPI inhibitors,
    which are typically larger and more lipophilic than inhibitors of standard binding sites. It has been published in:
    [Morelli X, Bourgeas R, Roche P. (2011) Chemical and structural lessons from recent successes in protein–protein interaction inhibition](https://pubmed.ncbi.nlm.nih.gov/21684802/)

    Also see: [Shin et al. (2020) Current Challenges and Opportunities in Designing Protein–Protein Interaction Targeted Drugs](https://pubmed.ncbi.nlm.nih.gov/33209039/)

    It computes: `MW >= 400 & logP >= 4  & RINGS >=4 & HBA >= 4`


    !!! warning
        Do not use this for small molecules that are not PPI inhibitors !

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_rings: precomputed number of rings in the molecules.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro4: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)

    return mw >= 400 and clogp >= 4 and n_rings >= 4 and n_hba >= 4


def rule_of_three(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """Compute the rule-of-3. The rule-of-three is a rule of thumb for molecular fragments (and not small molecules) published in:
    [Congreve M, Carr R, Murray C, Jhoti H. (2003) `A "rule of three" for fragment-based lead discovery?`](https://pubmed.ncbi.nlm.nih.gov/14554012/)

    It computes: `MW <= 300 & logP <= 3 & HBA <= 3 & HBD <= 3 & ROTBONDS <= 3`

    !!! note
        TPSA is not used in this version of the rule of three. Other version uses `TPSA <= 60 AND logP in [-3, 3]` in addition

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro3: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )

    return mw <= 300 and clogp <= 3 and n_hbd <= 3 and n_hba <= 3 and n_rotatable_bonds <= 3


def rule_of_three_extended(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """Compute the extended rule-of-3 which is an extension of the rule of three that
    includes and `TPSA` and relaxes `HBA` constraints.

    It computes: `MW <= 300 & logP in [-3, 3]  & HBA <= 6 & HBD <= 3 & ROTBONDS <= 3 & TPSA <= 60`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro3: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )

    return (
        mw <= 300
        and in_range(clogp, -3, 3)
        and n_hba <= 6
        and n_hbd <= 3
        and n_rotatable_bonds <= 3
        and tpsa <= 60
    )


def rule_of_two(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Computes rules-of-2 for reagent (building block design). It aims for prioritization of reagents that typically
    do not add more than 200 Da in MW or 2 units of clogP. The rule of two has been described in:

    [Goldberg et al. (2015) Designing novel building blocks is an overlooked strategy to improve compound quality](http://csmres.co.uk/cs.public.upd/article-downloads/Designing-novel-building-blocks.pdf)

    !!! note
        Their analysis showed that molecular weight (MW) and clogP were important factors in the frequency of use of reagents.
        Other parameters, such as TPSA, HBA, HBD and ROTBONDS count, were less important.

    It computes `MW <= 200 & logP <= 2 & HBA <= 4 & HBD <= 2`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ro2: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)

    return mw <= 200 and clogp <= 2 and n_hba <= 4 and n_hbd <= 2


def rule_of_ghose(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    mr: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the Ghose filter. The Ghose filter is a drug-like filter described in:
    [Ghose, AK.; Viswanadhan, VN.; Wendoloski JJ. (1999) A knowledge-based approach in designing combinatorial or medicinal
    chemistry libraries for drug discovery.1. A qualitative and quantitative characterization of known drug databases.](https://pubs.acs.org/doi/10.1021/cc9800071)

    It computes: `MW in [160, 480] & logP in [-0.4, 5.6] & Natoms in [20, 70] & refractivity in [40, 130]`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        mr: precomputed molecule refractivity.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        rog: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    num_atoms = mol.GetNumAtoms()  # ghose seems to use total number of atoms not heavy
    mr = mr if mr is not None else dm.descriptors.refractivity(mol)
    return (
        in_range(mw, 160, 480)
        and in_range(clogp, -0.4, 5.6)
        and in_range(num_atoms, 20, 70)
        and in_range(mr, 40, 130)
    )


def rule_of_veber(
    mol: Union[dm.Mol, str],
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the Veber filter. The Veber filter is a druglike filter for orally active drugs described in:

    [Veber et. al. (2002) Molecular Properties That Influence the Oral Bioavailability of Drug Candidates](https://pubmed.ncbi.nlm.nih.gov/12036371/)

    It computes: `ROTBONDS <= 10 & TPSA < 140`

    Args:
        mol: input molecule
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        rov: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)

    return n_rotatable_bonds <= 10 and tpsa <= 140


def rule_of_reos(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    charge: Optional[int] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_heavy_atoms: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the REOS filter. The REOS filter is a filter designed to filter out unuseful compounds from HTS screening results.
    The filter is described in: [Waters & Namchuk (2003) Designing screens: how to make your hits a hit.](https://pubmed.ncbi.nlm.nih.gov/12669025/)

    It computes: `MW in [200, 500] & logP in [-5, 5] & HBA in [0, 10] & HBD in [0, 5] & charge in [-2, 2] & ROTBONDS in [0, 8] & NHeavyAtoms in [15, 50]`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        charge: precomputed formal charge.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        n_heavy_atoms: precomputed number of heavy atoms in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        ror: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_heavy_atoms = n_heavy_atoms if n_heavy_atoms is not None else dm.descriptors.n_heavy_atoms(mol)
    charge = charge if charge is not None else dm.descriptors.formal_charge(mol)

    return (
        in_range(mw, 200, 500)
        and in_range(clogp, -5, 5)
        and in_range(n_hba, 0, 10)
        and in_range(n_hbd, 0, 5)
        and in_range(n_rotatable_bonds, 0, 8)
        and in_range(n_heavy_atoms, 15, 50)
    )


def rule_of_chemaxon_druglikeness(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the druglikeness filter according to chemaxon:

    It computes: `MW < 400 & logP < 5 & HBA <= 10 & HBD <= 5 & ROTBONDS < 5 & RINGS > 0`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        n_rings: precomputed number of rings in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        roc: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)

    return mw < 400 and clogp < 5 and n_hbd <= 5 and n_hba <= 10 and n_rotatable_bonds < 5 and n_rings > 0


def rule_of_egan(
    mol: Union[dm.Mol, str],
    clogp: Optional[float] = None,
    tpsa: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute passive intestinal absorption according to Egan Rules as described in:
    [Egan, William J., Kenneth M. Merz, and John J. Baldwin (2000) Prediction of drug absorption using multivariate statistics.](https://pubs.acs.org/doi/abs/10.1021/jm000292e)

    It computes: `TPSA in [0, 132] & logP in [-1, 6]`

    !!! abstract "Summary of the paper"
        The author built a multivariate statistics model of passive intestinal absorption with robust outlier detection.
        Outliers were identified as being actively transported. They chose PSA and AlogP98 (cLogP), based on consideration of the physical processes
        involved in membrane permeability and the interrelationships and redundancies between other available descriptors.

        **Compounds, which had been assayed for Caco-2 cell permeability, demonstrated a good rate of successful predictions (74−92%)**

    Args:
        mol: input molecule
        clogp: precomputed cLogP.
        tpsa: precomputed TPSA.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        roe: True if molecule is compliant, False otherwise
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    return in_range(tpsa, 0, 132) and in_range(clogp, -1, 6)


def rule_of_pfizer_3_75(
    mol: Union[dm.Mol, str],
    clogp: Optional[float] = None,
    tpsa: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute Pfizer 3/75 Rule for invivo toxicity. It has been described in:
    [Hughes, et al. (2008) Physiochemical drug properties associated with in vivo toxicological outcomes](https://pubmed.ncbi.nlm.nih.gov/18691886/)
    [Price et al. (2009) Physicochemical drug properties associated with in vivo toxicological outcomes: a review](https://pubmed.ncbi.nlm.nih.gov/19519283/)

    It computes: `TPSA >= 75 & logP <= 3`

    !!! abstract "Summary of the paper"
        * In vivo toleration (IVT) studies on 245 preclinical Pfizer compounds found an increased likelihood of toxic events for less polar, more lipophilic compounds.
        * Compounds with `low clogP / high TPSA` are ∼ 2.5 times more likely **to not have any toxity issue** at a fixed concentration of 10 uM (total) or 1 uM (free);
        * Compounds with `high clogP / low TPSA` are ∼ 2.5 times more likely **to have a toxicity finding**; this represents an overall odds >= 6.

    Args:
        mol: input molecule
        clogp: precomputed cLogP.
        tpsa: precomputed TPSA.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        rop: True if molecule is compliant, False otherwise

    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)

    return not (clogp > 3 and tpsa < 75)


def rule_of_gsk_4_400(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute GSK Rule (4/400) for druglikeness using interpretable ADMET rule of thumb. It has been described in:
    [Gleeson, M. Paul (2008). Generation of a set of simple, interpretable ADMET rules of thumb.](https://pubs.acs.org/doi/10.1021/jm701122q)

    It computes: `MW <= 400 & logP <= 4`.

    !!! abstract "Summary of the paper"
        - The rule are based on a set of consistent structure-property guides determined from an analysis of a number of key
            ADMET assays run within GSK: solubility, permeability, bioavailability, volume of distribution, plasma protein binding,
            CNS penetration, brain tissue binding, P-gp efflux, hERG inhibition, and cytochrome P450 1A2/2C9/2C19/2D6/3A4 inhibition.
        - Conclusion: _It is clear from the analyses reported herein that almost all ADMET parameters deteriorate with either increasing molecular weight,
            logP, or both, with ionization state playing either a beneficial or detrimental affect depending on the parameter in question._

    Args:
        mol: input molecule
        clogp: precomputed cLogP.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        rog: True if molecule is compliant, False otherwise

    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)

    return mw <= 400 and clogp <= 4


def rule_of_oprea(
    mol: Union[dm.Mol, str],
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Computes Oprea's rule of drug likeness obtained by comparing drug vs non drug compounds across multiple datasets.
    The rules have been described in: [Oprea (2000) Property distribution of drug-related chemical databases](https://pubmed.ncbi.nlm.nih.gov/10756480/)

    It computes: `HBD in [0, 2] & HBA in [2, 9] & ROTBONDS in [2,8] and RINGS in [1, 4]`

    !!! abstract "Summary of the paper"
        Seventy percent of the `drug-like` compounds were found between the following limits: `0 ≤ HDO ≤ 2, 2 ≤ HAC ≤ 9, 2 ≤ RTB ≤ 8, and 1 ≤ RNG ≤ 4`

    Args:
        mol: input molecule
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        n_rings: precomputed number of rings in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns
        roo: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)

    return (
        in_range(n_hbd, 0, 2)
        and in_range(n_hba, 2, 9)
        and in_range(n_rotatable_bonds, 2, 8)
        and in_range(n_rings, 1, 4)
    )


def rule_of_xu(
    mol: Union[dm.Mol, str],
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    n_heavy_atoms: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Computes Xu's rule of drug likeness as described in:
    [Xu & Stevenson (2000), Drug-like Index: A New Approach To Measure Drug-like Compounds and Their Diversity.](https://pubmed.ncbi.nlm.nih.gov/11045811/)

    It computes `HBD <= 5 & HBA <= 10 & ROTBONDS in [2, 35] & RINGS in [1, 7] & NHeavyAtoms in [10, 50]`.

    !!! note
        A compound's Drug Likeness Index is calculated based upon the knowledge derived from known drugs selected from Comprehensive Medicinal Chemistry (CMC) database.

    Args:
        mol: input molecule
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        n_rings: precomputed number of rings in the molecule.
        n_heavy_atoms: precomputed number of rings in the molecule.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns
        rox: True if molecule is compliant, False otherwise

    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)
    n_heavy_atoms = n_heavy_atoms if n_heavy_atoms is not None else dm.descriptors.n_heavy_atoms(mol)

    return (
        n_hba <= 10
        and n_hbd <= 5
        and in_range(n_rotatable_bonds, 2, 35)
        and in_range(n_rings, 1, 7)
        and in_range(n_heavy_atoms, 10, 50)
    )


def rule_of_cns(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Computes drug likeness rule for CNS penetrant molecules as described in:
    [Jeffrey & Summerfield (2010) Assessment of the blood-brain barrier in CNS drug discovery.](https://pubmed.ncbi.nlm.nih.gov/19664709/)

    It computes: `MW in [135, 582]  & logP in [-0.2, 6.1] & TPSA in [3, 118] & HBD <= 3 & HBA <= 5`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed logP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        roc: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)

    return (
        in_range(mw, 135, 582)
        and in_range(clogp, -0.2, 6.1)
        and in_range(tpsa, 3, 118)
        and n_hbd <= 3
        and n_hba <= 5
    )


def rule_of_respiratory(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_hba: Optional[float] = None,
    n_hbd: Optional[float] = None,
    tpsa: Optional[int] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_rings: Optional[int] = None,
    **kwargs: Any,
) -> bool:
    """
    Computes drug likeness rule for Respiratory (nasal/inhalatory) molecules as described in:
    [Ritchie et al. (2009) Analysis of the Calculated Physicochemical Properties of Respiratory Drugs: Can We Design for Inhaled Drugs Yet?](https://pubmed.ncbi.nlm.nih.gov/19275169/)

    It computes: `MW in [240, 520]  & logP in [-2, 4.7] & HBONDS in [6, 12] & TPSA in [51, 135] & ROTBONDS in [3,8] & RINGS in [1,5]`

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed logP.
        n_hba: precomputed number of HBA.
        n_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds in the molecule.
        n_rings: precomputed number of rings.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    Returns:
        roc: True if molecule is compliant, False otherwise
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_hbd = n_hbd if n_hbd is not None else dm.descriptors.n_hbd(mol)
    n_hba = n_hba if n_hba is not None else dm.descriptors.n_hba(mol)
    n_hbonds = n_hbd + n_hba
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_rings = n_rings if n_rings is not None else dm.descriptors.n_rings(mol)

    return (
        in_range(mw, 240, 520)
        and in_range(clogp, -2, 4.7)
        and in_range(n_hbonds, 6, 12)
        and in_range(tpsa, 51, 135)
        and in_range(n_rotatable_bonds, 3, 8)
        and in_range(n_rings, 1, 5)
    )


def rule_of_generative_design(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_lipinski_hba: Optional[float] = None,
    n_lipinski_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_hetero_atoms: Optional[int] = None,
    charge: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute druglikeness rule of generative design.

    This set of rules are proprietary to © Valence Discovery and have been curated to better prioritize molecules
    suggested by generative models for small molecules.

    It computes:

    ```
    MW in [200, 600] & logP < in [-3, 6] & HBD <= 7  & HBA <= 12 & TPSA in [40, 180] &
    ROTBONDS <= 15 & RIGID BONDS <= 30 & N_AROMATIC_RINGS <= 5 & N_FUSED_AROMATIC_RINGS_TOGETHER <= 2 &
    MAX_SIZE_RING_SYSTEM <= 18  & N_CARBONS in [3, 40] & N_HETEROATOMS in [1, 15] & CHARGE in [-2, 2] &
    N_ATOM_CHARGE <= 2 & N_TOTAL_ATOMS < 70 & N_HEAVY_METALS < 1
    ```

    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_lipinski_hba: precomputed number of HBA.
        n_lipinski_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        n_hetero_atoms: precomputed number of heteroatoms.
        charge: precomputed charge.
        **kwargs: Allow extra arguments for descriptors pre-computation.

    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)
    mol = dm.sanitize_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    mw = mw if mw is not None else dm.descriptors.mw(mol)
    clogp = clogp if clogp is not None else dm.descriptors.clogp(mol)
    n_lipinski_hba = n_lipinski_hba if n_lipinski_hba is not None else dm.descriptors.n_lipinski_hba(mol)
    n_lipinski_hbd = n_lipinski_hbd if n_lipinski_hbd is not None else dm.descriptors.n_lipinski_hbd(mol)
    tpsa = tpsa if tpsa is not None else dm.descriptors.tpsa(mol)
    n_rotatable_bonds = (
        n_rotatable_bonds if n_rotatable_bonds is not None else dm.descriptors.n_rotatable_bonds(mol)
    )
    n_hetero_atoms = n_hetero_atoms if n_hetero_atoms is not None else dm.descriptors.n_hetero_atoms(mol)

    # reionize first before computing charge
    with dm.without_rdkit_log():
        standard_mol = dm.standardize_mol(mol, reionize=True, uncharge=False, stereo=False)

    charge = charge if charge is not None else dm.descriptors.formal_charge(standard_mol)
    num_charged_atom = dm.descriptors.n_charged_atoms(standard_mol)
    n_rigid_bonds = dm.descriptors.n_rigid_bonds(mol)
    ring_system = dm.compute_ring_system(mol, include_spiro=False)
    max_size_ring = 0 if len(ring_system) == 0 else max([len(x) for x in ring_system])
    n_carbons = len([at for at in mol.GetAtoms() if at.GetSymbol() == "C"])
    n_aromatic_rings = dm.descriptors.n_aromatic_rings(mol)
    n_fused_aro_rings = n_fused_aromatic_rings(mol, require_all_aromatic=True, pairwise=False)
    n_total_atoms = mol.GetNumAtoms()
    n_heavy_mets = n_heavy_metals(mol)

    return (
        in_range(mw, 200, 600)
        and in_range(clogp, -3, 6)
        and n_lipinski_hbd <= 7
        and n_lipinski_hba <= 12
        and in_range(tpsa, 40, 180)
        and n_rotatable_bonds <= 15
        and n_rigid_bonds <= 30
        and n_aromatic_rings <= 5
        and n_fused_aro_rings <= 2
        and max_size_ring <= 18
        and in_range(n_carbons, 3, 40)
        and in_range(n_hetero_atoms, 1, 15)
        and in_range(charge, -2, 2)
        and num_charged_atom <= 2
        and n_total_atoms < 70
        and n_heavy_mets < 1
    )


def rule_of_generative_design_strict(
    mol: Union[dm.Mol, str],
    mw: Optional[float] = None,
    clogp: Optional[float] = None,
    n_lipinski_hba: Optional[float] = None,
    n_lipinski_hbd: Optional[float] = None,
    tpsa: Optional[float] = None,
    n_rotatable_bonds: Optional[int] = None,
    n_hetero_atoms: Optional[int] = None,
    charge: Optional[float] = None,
    **kwargs: Any,
) -> bool:
    """
    Compute the strict version of the druglikeness rule of generative design, which includes long aliphatic chains.

    This set of rules are proprietary to © Valence Discovery and have been curated to better prioritize molecules
    suggested by generative models for small molecules.

    It computes:

    ```
    MW in [200, 600] & logP < in [-3, 6] & HBD <= 7  & HBA <= 12 & TPSA in [40, 180] &
    ROTBONDS <= 15 & RIGID BONDS <= 30 & N_AROMATIC_RINGS <= 5 & N_FUSED_AROMATIC_RINGS_TOGETHER <= 2 &
    MAX_SIZE_RING_SYSTEM <= 18  & N_CARBONS in [3, 40] & N_HETEROATOMS in [1, 15] & CHARGE in [-2, 2] &
    N_ATOM_CHARGE <= 2 & N_TOTAL_ATOMS < 70 & N_HEAVY_METALS < 1 & N_STEREO_CENTER <= 3 &
    HAS_NO_SPIDER_SIDE_CHAINS & FRACTION_RING_SYSTEM >= 0.25
    ```

    By default SPIDER_SIDE_CHAINS are defined as having at least 2 'chains' of  >=4 consecutif atoms in side chains (not part of any ring system)


    Args:
        mol: input molecule
        mw: precomputed molecular weight.
        clogp: precomputed cLogP.
        n_lipinski_hba: precomputed number of HBA.
        n_lipinski_hbd: precomputed number of HBD.
        tpsa: precomputed TPSA.
        n_rotatable_bonds: precomputed number of rotatable bonds.
        n_hetero_atoms: precomputed number of heteroatoms.
        charge: precomputed charge.
        **kwargs: Allow extra arguments for descriptors pre-computation.
    """

    mol = dm.to_mol(mol)
    mol = dm.sanitize_mol(mol)

    if mol is None:
        raise ValueError("Molecule is None")

    generative_rule = rule_of_generative_design(
        mol=mol,
        mw=mw,
        clogp=clogp,
        n_lipinski_hba=n_lipinski_hba,
        n_lipinski_hbd=n_lipinski_hbd,
        tpsa=tpsa,
        n_rotatable_bonds=n_rotatable_bonds,
        n_hetero_atoms=n_hetero_atoms,
        charge=charge,
        **kwargs,
    )

    n_stereo_center = dm.descriptors.n_stereo_centers(mol)
    has_spider_appendages = has_spider_chains(mol)
    good_fraction_ring_system = fraction_atom_in_scaff(mol) >= 0.25

    return (
        generative_rule and good_fraction_ring_system and not has_spider_appendages and n_stereo_center <= 3
    )



from ._utils import in_range
from ._utils import n_heavy_metals
from ._utils import has_spider_chains
from ._utils import n_fused_aromatic_rings
from ._utils import fraction_atom_in_scaff
from ._utils import list_descriptors

from . import basic_rules

from ._rule_filter import RuleFilters



from typing import Callable

import pytest

import pandas as pd
import datamol as dm
import numpy as np

import medchem as mc


def test_list_descriptors():
    list_descs = mc.rules.list_descriptors()
    assert isinstance(list_descs, list)
    assert len(list_descs) > 0
    assert all(isinstance(x, str) for x in list_descs)


def test_has_spider_chains():
    mols_with_appendages = [
        "CCNCC1CCC(CNC(C)C)C1",
        "CCCCC1=CN(C=C1CCC)C1=CC(=CC=C1)C(C)CCO",
        "CCCCC1=CC2=C(OC(CN(C)CC)O2)C=C1",
    ]
    mols_without_appendages = [
        "CC(C)C(O)C(O)C(N)=O",
        "OC1=CC=CC(CC2CC2)=C1",
        "CCN(C)C1=CC=CC(CC(O)CC(C)O)=C1",
        "CN1CCC(CC(=O)NCC2=CC=CC=C2)C1",
        "CCCCC1=CC(=CC=C1)C(CC)C(C)C",
    ]

    assert all(
        mc.rules.has_spider_chains(dm.to_mol(x)) for x in mols_with_appendages
    ), "Fail appendage test for mols with appendages"
    assert all(
        not mc.rules.has_spider_chains(dm.to_mol(x)) for x in mols_without_appendages
    ), "Fail appendage test for mols without appendages"


def test_fraction_ring_system():
    mols = [
        "CCNCC1CCC(CNC(C)C)C1",
        "CCCCC1=CN(C=C1CCC)C1=CC(=CC=C1)C(C)CCO",
        "CCCCC1=CN(C=C1CCC)C1=CC(=CC=C1)C(C)CCCCC1CCO1",
        "CC(C)C(O)C(O)C(N)=O",
        "C1CCCCCC1",
    ]
    mols = [dm.to_mol(x) for x in mols]
    expected_results = [5 / 14, 11 / 23, 20 / 28, 0, 1]  # n_atoms_rings / n_atoms
    np.testing.assert_allclose(expected_results, [mc.rules.fraction_atom_in_scaff(x) for x in mols])


def test_n_fused_aromatic_rings():
    smiles = [
        "C1CNC2CCCNC2C1",
        "C1=CC2=C(C=C1)C1=C(C=C2)N=CC=C1",
        "C1CC1C1=CN=CC(=C1)C1=CC2=C(OCO2)C=C1",
        "C1OC2=C(O1)C=C(C=C2)C1=CN=C2C=CC=CC2=C1",
        "O=C(NC1=CC=C2N=CNC2=C1)C1=CC=C2C=CC=CC2=C1",
    ]
    mols = [dm.to_mol(x) for x in smiles]
    expected_results_pairwise = [0, 2, 0, 1, 2]
    expected_results = [0, 1, 0, 1, 2]
    expected_results_no_all_aro = [0, 1, 1, 2, 2]
    results = [mc.rules.n_fused_aromatic_rings(x) for x in mols]
    results_pairwise = [mc.rules.n_fused_aromatic_rings(x, pairwise=True) for x in mols]
    results_no_all_aro = [mc.rules.n_fused_aromatic_rings(x, require_all_aromatic=False) for x in mols]
    assert results == expected_results, "Fail fused aromatic ring test"
    assert expected_results_pairwise == results_pairwise, "Fail pairwise fused aromatic ring test"
    assert results_no_all_aro == expected_results_no_all_aro, "Fail not fully aromatic ring test"


def test_n_heavy_metals():
    smiles = [
        r"[Pd++].[H][C@]1(C)\C2=C\C3=C(C(C)=O)C(C)=C([N-]3)\C=C3/N=C(/C(/CC(=O)OC)=C4\[N-]\C(=C/C(=N2)[C@]1([H])CC)C(C)=C4C(=O)NCCS(O)(=O)=O)[C@@]([H])(CCC(O)=O)[C@]3([H])C",
        "[Na+].[Na+].[Na+].OC(CC([O-])=O)(CC([O-])=O)C([O-])=O",
        "CC(=O)OC1=CC=CC=C1C(O)=O",
    ]
    mols = [dm.to_mol(x) for x in smiles]
    output = [mc.rules.n_heavy_metals(m) for m in mols]
    expected_output = [1, 0, 0]
    assert output == expected_output, "Fail heavy atom test"


def test_rule_of_generative_stereo_case():
    smiles = [
        "CCC(O)C1=CC=C(C=C1)S(=O)(=O)C1CCNC(C)C1",
        "CC(F)C(O)C1=CC=C(C=C1)S(=O)(=O)C1CCNC(C)C1",
    ]
    mols = [dm.to_mol(x) for x in smiles]
    output = [mc.rules.basic_rules.rule_of_generative_design(x) for x in mols]
    excepted_output = [True, True]
    assert output == excepted_output, "Fail simple rule of generative"

    output = [mc.rules.basic_rules.rule_of_generative_design_strict(x) for x in mols]
    excepted_output = [True, False]  # second molecule fails due to n_stereo_center
    assert output == excepted_output, "Fail rule of generative stereo center test"


def test_basic_rules():
    data = dm.data.freesolv()
    data = data.iloc[:50]

    all_basic_rules = mc.rules.RuleFilters.list_available_rules()["name"].values
    for rule_name in all_basic_rules:
        rule_fn = getattr(mc.rules.basic_rules, rule_name)
        rule_fn(data["smiles"].values[10])


def test_rfilter():
    data = dm.data.freesolv()
    data = data.iloc[:50]

    def my_rules(mol, **kwargs):
        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Molecule is None")

        return mol.GetNumHeavyAtoms() / dm.descriptors.clogp(mol) > 2

    rfilter = mc.rules.RuleFilters(
        rule_list=[my_rules, "rule_of_five"],
        rule_list_names=["my_rules", "rule_of_five"],
    )
    out = rfilter(data["smiles"].tolist())
    assert out.shape[-1] == 5
    assert out.shape[0] == 50
    assert out["my_rules"].sum() == 45
    assert out["rule_of_five"].sum() == 48
    assert out["pass_all"].sum() == 44
    assert out["pass_any"].sum() == 49


def test_rule_basic_raise_error_if_none():
    all_basic_rules = mc.rules.RuleFilters.list_available_rules()["name"].values
    for rule_name in all_basic_rules:
        rule_fn = getattr(mc.rules.basic_rules, rule_name)

        with pytest.raises(ValueError):
            rule_fn(None)


def test_rule_filters_no_names():
    def my_rules(mol, **kwargs):
        return True

    rfilter = mc.rules.RuleFilters(rule_list=[my_rules, "rule_of_five"])

    assert list(rfilter.rules.keys()) == ["my_rules", "rule_of_five"]


def test_rule_filters_class():
    rfilter = mc.rules.RuleFilters(rule_list=["rule_of_three", "rule_of_five"])
    assert len(rfilter) == 2
    assert len(rfilter.rules) == 2
    assert isinstance(rfilter.rules["rule_of_three"], Callable)


def test_rule_filters_invalid():
    rfilter = mc.rules.RuleFilters(rule_list=["rule_of_five"])

    with pytest.raises(ValueError):
        rfilter(mols=[None, "CCCCO"], fail_if_invalid=True)

    results = rfilter(mols=[None, "CCCCO"], fail_if_invalid=False)
    assert results["mol"].isnull().sum() == 1

    assert pd.isna(results.iloc[0]["pass_all"])
    assert pd.isna(results.iloc[0]["pass_any"])
    assert pd.isna(results.iloc[0]["rule_of_five"])

    assert results.iloc[1]["pass_all"] is True
    assert results.iloc[1]["pass_any"] is True
    assert results.iloc[1]["rule_of_five"] is True


def test_rule_filters_list():
    ll = mc.rules.RuleFilters.list_available_rules()
    assert isinstance(ll, pd.DataFrame)
    assert "name" in ll.columns

    ll = mc.rules.RuleFilters.list_available_rules_names()
    assert isinstance(ll, list)
    assert all(isinstance(x, str) for x in ll)
    assert len(ll) > 0

    ll = mc.rules.RuleFilters.list_available_rules("building block")
    assert isinstance(ll, pd.DataFrame)
    assert "name" in ll.columns

    ll = mc.rules.RuleFilters.list_available_rules_names("building block", "fragment")
    assert isinstance(ll, list)
    assert all(isinstance(x, str) for x in ll)
    assert len(ll) > 0



import datamol as dm
import pandas as pd

from medchem.utils.loader import get_data_path
from medchem.catalogs import merge_catalogs
from medchem.catalogs import NamedCatalogs
from medchem.catalogs import catalog_from_smarts


def test_catalog_merge():
    data = dm.data.freesolv()
    mols = data["smiles"].apply(dm.to_mol).tolist()

    tox = NamedCatalogs.tox(pains_a=False, pains_b=False, pains_c=False, nih=True, brenk=True)
    nih = NamedCatalogs.nih()
    brenk = NamedCatalogs.brenk()

    merged_tox = merge_catalogs(nih, brenk)

    toxic1 = [tox.HasMatch(m) for m in mols]
    toxic2 = [merged_tox.HasMatch(m) for m in mols]
    assert toxic1 == toxic2


def test_catalog_from_smarts():
    smarts_bank = pd.read_csv(get_data_path("smarts_bank.csv"))
    custom_catalog = catalog_from_smarts(
        smarts_bank["smarts"],
        smarts_bank["name"],
        entry_as_inds=False,
    )

    assert custom_catalog.GetNumEntries() == 30


def test_NamedCatalogs():
    for catalog_name in [
        "alarm_nmr",
        "alerts",
        "alphascreen",
        "bms",
        "bredt",
        "brenk",
        "carcinogen",
        "chelator",
        "chemical_groups",
        "dnabinder",
        "dundee",
        "electrophilic",
        "glaxo",
        "gst_hitters",
        "his_hitters",
        "hitters",
        "inpharmatica",
        "ld50_oral",
        "lint",
        "luciferase",
        "mlsmr",
        "nibr",
        "nih",
        "pains",
        "pains_a",
        "pains_b",
        "pains_c",
        "reactive_unstable_toxic",
        "schembl",
        "skin",
        "tox",
        "toxicophore",
        "unstable_graph",
        "zinc",
    ]:
        catalog = getattr(NamedCatalogs, catalog_name)()
        assert catalog.GetNumEntries() > 0



def test_import():
    import medchem  # noqa: F401
    import medchem.utils  # noqa: F401
    import medchem.groups  # noqa: F401
    import medchem.catalogs  # noqa: F401
    import medchem.constraints  # noqa: F401
    import medchem.complexity  # noqa: F401
    import medchem.rules  # noqa: F401
